#include "TArtCalibFocalPlane.h" 
#include "TArtPPAC.h"
#include "TArtFocalPlane.h"
#include "TArtFocalPlanePara.h"
#include "TArtBigRIPSParameters.h"
#include "TArtCore.h" 
#include "TArtStoreManager.h"

#include <TROOT.h>
#include <TMath.h>
#include <TClonesArray.h>

//__________________________________________________________
TArtCalibFocalPlane::TArtCalibFocalPlane()
  :fDataLoaded(false), fReconstructed(false)
{
  TArtCore::Info(__FILE__,"Creating the FocalPlane objects...");
  sman = TArtStoreManager::Instance();
  setup = (TArtBigRIPSParameters*)sman->FindParameters("BigRIPSParameters");
  Initialize();
}

//__________________________________________________________
TArtCalibFocalPlane::~TArtCalibFocalPlane()  {
  ClearData();
}

//__________________________________________________________
void TArtCalibFocalPlane::Initialize()   {
  fFocalPlaneArray = new TClonesArray("TArtFocalPlane",10);
  fFocalPlaneArray->SetName("BigRIPSFocalPlane");
  fFocalPlaneArray->SetOwner();
  sman->AddDataContainer(fFocalPlaneArray);
}

//__________________________________________________________
void TArtCalibFocalPlane::LoadData(){

  TClonesArray *ppac_array = (TClonesArray *)sman->FindDataContainer("BigRIPSPPAC");

  Int_t num_ppac = ppac_array->GetEntries();
  char name[64];

  for(int i=0;i<num_ppac;i++){
    TArtPPAC *ppac = (TArtPPAC *)ppac_array->At(i);
    Int_t ppac_fpl = ppac->GetFpl();
    TArtFocalPlane * fpl = FindFocalPlane(ppac_fpl);
    if(NULL==fpl){
      Int_t nfpl=GetNumFocalPlane();
      new ((*fFocalPlaneArray)[nfpl]) TArtFocalPlane();
      fpl = (TArtFocalPlane *)fFocalPlaneArray->At(nfpl);
      TArtFocalPlanePara * para = setup->GetFocalPlanePara(ppac_fpl);
      sprintf(name,"Fpl%d",para->GetFpl());
      fpl->SetDetectorName(name);
      fpl->SetFpl(para->GetFpl());
      fpl->SetZpos(para->GetFocusPos());

      std::vector <TArtPPAC*> * tmp_ppacs = new std::vector <TArtPPAC*>;
      tmp_ppacs->push_back(ppac);
      fPPACArrayBuffer.push_back(tmp_ppacs);
    }
    else{
      for(int j=0;j<fPPACArrayBuffer.size();j++) 
	if(ppac_fpl == (*fPPACArrayBuffer[j])[0]->GetFpl()){
	  fPPACArrayBuffer[j]->push_back(ppac);
	  break;
	}
    }
  }

  fDataLoaded = true;
  return;
}

//__________________________________________________________
void TArtCalibFocalPlane::ClearData()   {
  fFocalPlaneArray->Clear();
  for(int i=0;i<fPPACArrayBuffer.size();i++){
    fPPACArrayBuffer[i]->clear();
    delete fPPACArrayBuffer[i];
  }
  fPPACArrayBuffer.clear();
  fDataLoaded = false;
  fReconstructed = false;
  return;
}

//__________________________________________________________
void TArtCalibFocalPlane::ReconstructData()   { // call after the raw data are loaded

  if(!fDataLoaded) LoadData();

  //on the assumption of linear track, the track is reconstructed analytically.
  //X = a + b * Z;

  for(int i=0;i<GetNumFocalPlane();i++){

    TArtFocalPlane *fpl = (TArtFocalPlane *)fFocalPlaneArray->At(i);
    TVectorD * opt_vector = (TVectorD *)fpl->GetOptVector();
    std::vector<TArtPPAC *> ppac_array = *(fPPACArrayBuffer[i]);

    TMatrixD xvec(2,1); xvec.Zero();
    TMatrixD yvec(2,1); yvec.Zero();
    TMatrixD xmat(2,2); xmat.Zero();
    TMatrixD ymat(2,2); ymat.Zero();

    Double_t sum_zx[2] = {0,0};
    Double_t sum_x[2] = {0,0};
    Double_t sum_z[2] = {0,0};
    Double_t sum_zz[2] = {0,0};
    Double_t x, y, z_x, z_y;

    int nppac = ppac_array.size();
    Double_t pZ_f = fpl->GetZpos();
    TArtPPAC *ppac = NULL;

    int nfired_ppacx = 0;
    int nfired_ppacy = 0;

    if(nppac<=0){
    }
    else if(nppac==1) {
      ppac = ppac_array[0];
      (*opt_vector)(0) = ppac->GetX();
      (*opt_vector)(2) = ppac->GetY();
      if(ppac->IsFiredX()) nfired_ppacx = 1;
      if(ppac->IsFiredY()) nfired_ppacy = 1;
    }
    else { 
      for(int i=0;i<nppac;i++){
	ppac = ppac_array[i];
	x = ppac->GetX();
	y = ppac->GetY();
	z_x = ppac->GetXZPos();
	z_y = ppac->GetYZPos();
	if(ppac->IsFiredX()){
	xvec(0,0) += z_x * x; // b(1) in rayfit
	xvec(1,0) += x; // b(2) in rayfit
	xmat(0,1) += z_x; // a(1,2) in rayfit
	xmat(1,0) += z_x; // a(1,2) in rayfit
	xmat(0,0) += z_x * z_x;   // a(1,1) in rayfit
	xmat(1,1) ++;  // a(2,2) in rayfit
	nfired_ppacx ++;  // a(2,2) in rayfit
	}
	if(ppac->IsFiredY()){
	  yvec(0,0) += z_y * y; // b(1) in rayfit
	  yvec(1,0) += y; // b(2) in rayfit
	  ymat(0,1) += z_y; // a(1,2) in rayfit
	  ymat(1,0) += z_y; // a(1,2) in rayfit
	  ymat(0,0) += z_y * z_y;   // a(1,1) in rayfit
	  ymat(1,1) ++;  // a(2,2) in rayfit
	  nfired_ppacy ++;  // a(2,2) in rayfit
	}
      }

      // determine a and b in x = a + b * z (i.e.); 
      if(nfired_ppacx>1){
	TMatrixD rxvec = xmat.Invert() * xvec;
	(*opt_vector)(0) = rxvec(1,0);
	(*opt_vector)(1) = TMath::ATan(rxvec(0,0)) * 1000;
      }
      if(nfired_ppacy>1){
	TMatrixD ryvec = ymat.Invert() * yvec;
	(*opt_vector)(2) = ryvec(1,0);
	(*opt_vector)(3) = TMath::ATan(ryvec(0,0)) * 1000; // the value is given in mrad.;
      }

    } 

    fpl->SetNumFiredPPACX(nfired_ppacx);
    fpl->SetNumFiredPPACY(nfired_ppacy);
    
  }

  fReconstructed = true;
  return;

}

//__________________________________________________________
TArtFocalPlane * TArtCalibFocalPlane::GetFocalPlane(int i) {
  return (TArtFocalPlane *)fFocalPlaneArray->At(i);
}
//__________________________________________________________
Int_t TArtCalibFocalPlane::GetNumFocalPlane() {
  return fFocalPlaneArray->GetEntries();
}
//__________________________________________________________
TArtFocalPlane * TArtCalibFocalPlane::FindFocalPlane(Int_t fpl){
  for(int i=0;i<GetNumFocalPlane();i++)
    if(fpl == ((TArtFocalPlane*)fFocalPlaneArray->At(i))->GetFpl())
      return (TArtFocalPlane*)fFocalPlaneArray->At(i);
  return NULL;
}
