#include "TArtCalibNEBULA.hh" 
#include "TArtNEBULAPla.hh"
#include "TArtNEBULAPlaPara.hh"
#include "TArtNEBULAParameters.hh"
#include "TArtRawEventObject.h"
#include "TArtCore.h" 
#include "TArtStoreManager.h"
#include "TArtMapper.h"

#include <TClonesArray.h>
#include <iostream>
#include <cmath>

//__________________________________________________________
TArtCalibNEBULA::TArtCalibNEBULA() : fReconstructed(false), fNEBULAParameters(0)
{
  TArtCore::Info(__FILE__,"Creating the NEBULA detector objects...");

  fStoreManager = TArtStoreManager::Instance();
  fNEBULAPlaArray = new TArtNEBULAPlaArray("TArtNEBULAPla", 300);
  fNEBULAPlaArray->SetOwner();
  fNEBULAPlaArray->SetName("NEBULAPla");
  fNEBULAPlaParaArray = new TArtNEBULAPlaParaArray("TArtNEBULAPlaPara", 300);
  fNEBULAPlaParaArray->SetOwner();
  fNEBULAPlaParaArray->SetName("NEBULAPlaPara");
  fStoreManager->AddDataContainer(fNEBULAPlaArray);
  fNEBULAPlaParaArray = new TArtNEBULAPlaParaArray;
  fNEBULAParameters = (const TArtNEBULAParameters*)fStoreManager->FindParameters("NEBULAParameters");
  if(!fNEBULAParameters){
    TArtCore::Info(__FILE__,"Could not find NEBULAParameters.");
    throw;
  }
}

//__________________________________________________________
TArtCalibNEBULA::~TArtCalibNEBULA()
{
  delete fNEBULAPlaParaArray;
  delete fNEBULAPlaArray;
}

//__________________________________________________________
void TArtCalibNEBULA::LoadRawData()
{
  TArtRawEventObject* fEvent = (TArtRawEventObject*)fStoreManager->FindDataContainer("RawEvent");
  ClearData();

  for(int i=0; i<fEvent->GetNumSeg(); ++i){
    TArtRawSegmentObject* seg = fEvent->GetSegment(i);
    int device = seg->GetDevice();
    if(SAMURAI == device){
      for(int j=0; j<seg->GetNumData(); ++j){
	int fpl = seg->GetFP();
	int detector = seg->GetDetector();
	TArtRawDataObject* d = seg->GetData(j);
	int geo = d->GetGeo(); 
	int ch = d->GetCh(); 
	unsigned int val = d->GetVal();
	TArtRIDFMap mm(fpl,detector,geo,ch);
	const TArtNEBULAPlaPara* para = fNEBULAParameters->FindNEBULAPlaPara(mm);
	if(!para){
	  TArtCore::Info(__FILE__,"Could not find TArtNEBULAPara...: Dev:%d, %s",
			 device, mm.GetMapInfo());
	  continue;
	}
	if(para->GetIgnore()){
	  TArtCore::Debug(__FILE__,"Ignore TArtNEBULAPara...: Dev:%d, %s", device, mm.GetMapInfo());
	  continue;
	}

	if(FindNEBULAPlaPara(para->GetID(), para->GetUDLR())){
	  int npara = fNEBULAPlaParaArray->GetEntries();
	  new ((*fNEBULAPlaParaArray)[npara]) TArtNEBULAPlaPara(*para);
	}
	TArtNEBULAPla* pla = FindNEBULAPla(para->GetID());
	if(!pla){
	  int npla = fNEBULAPlaArray->GetEntries();
	  pla = new ((*fNEBULAPlaArray)[npla]) TArtNEBULAPla();
	  pla->SetID(para->GetID());
	  pla->SetFPl(para->GetFPl());
	  pla->SetDetectorName(para->GetDetectorName());
	  pla->SetLayer(para->GetLayer());
	  pla->SetSubLayer(para->GetSubLayer());
	}

	TString udlr = para->GetUDLR();
	// if(detector == NEBULAT)
	if(detector == 41){
	  if(udlr == "U"){
	    pla->SetTURaw(val);
	  }else if(udlr == "D"){
	    pla->SetTDRaw(val);
	  }
	}
	// if(detector == NEBULAQ)
	if(detector == 42){
	  if(udlr == "U"){
	    pla->SetQURaw(val);
	  }else if(udlr == "D"){
	    pla->SetQDRaw(val);
	  }
	}
      }
    }
  }

  return;
}

//__________________________________________________________
void TArtCalibNEBULA::ReconstructData()
{
  //共通のoffset。どう値を取ってくるかは後で考える(とりあえず手打ち？)。
  Double_t posxoff = 0;
  Double_t posyoff = 0;
  Double_t poszoff = 0;

  for(int i=0;i<GetNumNEBULAPla();++i){
    TArtNEBULAPla* pla = GetNEBULAPla(i);
    Int_t id = pla->GetID();
    const TArtNEBULAPlaPara* parau = fNEBULAParameters->FindNEBULAPlaPara(id, "U");
    if(!parau){
      TArtCore::Info(__FILE__,"cannot find para %d U", id);
      continue;
    }
    const TArtNEBULAPlaPara* parad = fNEBULAParameters->FindNEBULAPlaPara(id, "D");
    if(!parad){
      TArtCore::Info(__FILE__,"cannot find para %d D", id);
      continue;
    }
    //左右共通のparameterはparaに付け替えておく。
    const TArtNEBULAPlaPara* para = parau;

    int turaw = pla->GetTURaw();
    int tdraw = pla->GetTDRaw();
    int quraw = pla->GetQURaw();
    int qdraw = pla->GetQDRaw();

    int ihitmin = 0;
    if(turaw>0 && turaw<=4096) ++ihitmin;
    if(tdraw>0 && tdraw<=4096) ++ihitmin;      
    if(quraw>0 && quraw<=4096) ++ihitmin;
    if(qdraw>0 && qdraw<=4096) ++ihitmin;
    pla->SetIHitMin(ihitmin);

    if(ihitmin > 0){
      pla->SetDataState(1);
    }else{
      pla->SetDataState(-1);
      continue;
    }

    Double_t quped = quraw - parau->GetQPed();
    Double_t qdped = qdraw - parad->GetQPed();
    Double_t qucal = quped * parau->GetQCal();
    Double_t qdcal = qdped * parad->GetQCal();
    Double_t qaveped = sqrt(quped*qdped);
    Double_t qavecal = sqrt(qucal*qdcal);
    Double_t logqped = log(quped/qdped);
    Double_t logqcal = log(qucal/qdcal);

    //    Double_t tucal = turaw * parau->GetTCal() + parau->GetTOff();
    //    Double_t tdcal = tdraw * parad->GetTCal() + parad->GetTOff();
    Double_t tucal = turaw * parau->GetTCal();
    Double_t tdcal = tdraw * parad->GetTCal();
    Double_t tuslw = tucal - parau->GetTSlw()/sqrt(quped);
    Double_t tdslw = tdcal - parad->GetTSlw()/sqrt(qdped);
    Double_t    dtraw = turaw - tdraw;
    Double_t dtcal = tucal - tdcal;
    Double_t dtslw = tuslw - tdslw;
    Double_t taveraw = (turaw + tdraw)/2.;
    Double_t tavecal = (tucal + tdcal)/2 + para->GetTOff();
    Double_t taveslw = (tuslw + tdslw)/2 + para->GetTOff();

    Double_t poscal = dtcal * para->GetDTCal() + para->GetDTOff();
    Double_t posslw = dtslw * para->GetDTCal() + para->GetDTOff();

    Double_t pos[3];
    pos[0] =          para->GetPos(0) + posxoff;
    pos[1] = posslw + para->GetPos(1) + posyoff; 
    pos[2] =          para->GetPos(2) + poszoff;

    pla->SetQUPed(quped);
    pla->SetQDPed(qdped);
    pla->SetQUCal(qucal);
    pla->SetQDCal(qdcal);
    pla->SetQAvePed(qaveped);
    pla->SetQAveCal(qavecal);
    pla->SetLogQPed(logqped);
    pla->SetLogQCal(logqcal);

    pla->SetTUCal(tucal);
    pla->SetTDCal(tdcal);
    pla->SetTUSlw(tuslw);
    pla->SetTDSlw(tdslw);
    pla->SetDTRaw(dtraw);
    pla->SetDTCal(dtcal);
    pla->SetDTSlw(dtslw);    
    pla->SetTAveRaw(taveraw);
    pla->SetTAveCal(tavecal);
    pla->SetTAveSlw(taveslw);    

    pla->SetPosCal(poscal);
    pla->SetPosSlw(posslw);

    pla->SetTime(taveslw);
    pla->SetEnergy(qavecal);
    pla->SetPos(pos);
  }
  
  fReconstructed = true;
  return;
}

//__________________________________________________________
Int_t TArtCalibNEBULA::GetNumNEBULAPla() const
{
  return fNEBULAPlaArray->GetEntries();
}

//__________________________________________________________
TArtNEBULAPla* TArtCalibNEBULA::GetNEBULAPla(int i) const
{
  return (TArtNEBULAPla*)fNEBULAPlaArray->At(i);
}

//__________________________________________________________
TArtNEBULAPla* TArtCalibNEBULA::FindNEBULAPla(Int_t id) const
{
  for(int i=0; i<GetNumNEBULAPla(); ++i){
    TArtNEBULAPla* pla = GetNEBULAPla(i);
    if(id == pla->GetID()) return pla;
  }
  return 0;
}

//__________________________________________________________
TArtNEBULAPlaArray* TArtCalibNEBULA::GetNEBULAPlaArray() const
{
  return fNEBULAPlaArray;
}

//__________________________________________________________
Int_t TArtCalibNEBULA::GetNumNEBULAPlaPara() const
{
  return fNEBULAPlaParaArray->GetEntries();
}

//__________________________________________________________
TArtNEBULAPlaPara* TArtCalibNEBULA::GetNEBULAPlaPara(int i) const
{
  return (TArtNEBULAPlaPara*)fNEBULAPlaParaArray->At(i);
}

//__________________________________________________________
TArtNEBULAPlaPara* TArtCalibNEBULA::FindNEBULAPlaPara(int id, TString udlr) const
{
  for(int i=0; i<GetNumNEBULAPlaPara(); ++i){
    TArtNEBULAPlaPara* para = GetNEBULAPlaPara(i);
    if(id == para->GetID() && udlr == para->GetUDLR()) return para;
  }

  return 0;
}

//__________________________________________________________
TArtNEBULAPlaParaArray* TArtCalibNEBULA::GetNEBULAPlaParaArray() const
{
  return fNEBULAPlaParaArray;
}

//__________________________________________________________
void TArtCalibNEBULA::ClearData()
{
  fNEBULAPlaArray->Clear();
  fNEBULAPlaParaArray->Clear();
  fReconstructed = false;
  return;
}

